static char *SCCS_Id = "@(#)cache.c\tver 1.2 (92/06/04 11:08:35)";

/*
 * Page caching layer.
 *
 * Copyright 1992 Barry Brachman
 * See included file "copyright.h" for complete copyright information.
 */

#include "local.h"

static int getPage PROTO((int fd, PhysPagenum pagenum, int allocsize,
						  int nbytes, PageCache *cache));
#ifdef NOTDEF
static int putCache PROTO((int fd, int allocsize, PageCache *cache));
#endif NOTDEF

/*
 * Allocate and initialize a pointer to a new cache entry.
 */
DbmRc
CacheAlloc PARAMS((pagenum, nbytes, cache),
PhysPagenum pagenum		_AND_
int nbytes				_AND_
PageCache **cache)
{
	PageCache *pc;

    if ((pc = (PageCache *) NAMEDMALLOC("PageCache", sizeof(PageCache)))
		== NULL)
        return(DBM_MALLOC_FAILED);

	if ((pc->page = (void *) NAMEDMALLOC("CacheAlloc", nbytes)) == NULL) {
		FREE(pc);
		return(DBM_MALLOC_FAILED);
	}

	pc->pagenum = pagenum;
	pc->nbytes = nbytes;
	pc->flags = PAGE_CLEAN;
	pc->next = NULL;

	*cache = pc;
	return(DBM_OK);
}

void
CacheFree PARAMS((cache),
PageCache *cache)
{

	if (cache->page) {
		FREE(cache->page);
		cache->page = NULL;
	}
	cache->next = NULL;
	FREE(cache);
}

/*
 * Read PHYSPAGE.
 */
DbmRc
CacheRead PARAMS((trans, physpage, nbytes, cachep),
DbmTrans *trans				_AND_
PhysPagenum physpage		_AND_
int nbytes					_AND_
PageCache **cachep)
{
	DbmFile *dbf;
	DbmRc rc;
	PageCache *cache;

	if ((rc = CacheAlloc(physpage, nbytes, &cache)) != DBM_OK)
		return(rc);

	dbf = DBF(trans);
	if (!dbf->persistent) {
		bzero((char *) cache->page, nbytes);
		cache->flags = PAGE_NEW | PAGE_DIRTY;
	}
	else {
		if (getPage(dbf->fd, physpage, dbf->allocsize, nbytes, cache) < 0) {
			CacheFree(cache);
			return(DBM_PAGE_IO_ERROR);
		}
	}

	*cachep = cache;
	return(DBM_OK);
}

/*
 * Read one page, starting at PAGENUM, from file FD into CACHE.
 * Each page of the file is PAGESIZE bytes long.
 * Note that if a read returns 0, then EOF has been reached and the
 * buffer is zero filled.
 * Return 0 on success, -1 if I/O error.
 */
static int
getPage PARAMS((fd, pagenum, allocsize, nbytes, cache),
int fd						_AND_
PhysPagenum pagenum			_AND_
int allocsize				_AND_
int nbytes					_AND_
PageCache *cache)
{
	int n;

	n = SeekNRead(fd, cache->page, nbytes, pagenum * allocsize, SEEK_SET);
	if (n != nbytes) {
		if (n < 0)
			return(-1);
		bzero((char *) cache->page, nbytes);
		cache->flags = PAGE_NEW | PAGE_DIRTY;
	}
	else
		cache->flags = PAGE_CLEAN;

	cache->pagenum = pagenum;
	return(0);
}

/*
 */
DbmRc
CacheWrite PARAMS((trans, cache),
DbmTrans *trans			_AND_
PageCache *cache)
{

	if (cache->pagenum >=
		NPHYSMAP_PAGES * DBF(trans)->pagesize / sizeof(PhysPagenum)) {
		fprintf(stderr, "CacheWrite: pagenumber too big!\n");
		return(DBM_PAGENUMBER_TOO_BIG);
	}

	return(DBM_OK);
}

#ifdef NOTDEF
/*
 * Write a page into file FD from CACHE.
 * Each page of the file is ALLOCSIZE bytes long.
 * Return 0 on success, -1 if I/O error.
 */
static int
putCache(int fd, int allocsize, PageCache *cache)
{

	if (cache->flags & PAGE_DIRTY) {
		if (SeekNWrite(fd, cache->page, cache->nbytes,
					   cache->pagenum * allocsize, SEEK_SET) != cache->nbytes)
			return(-1);
		cache->flags = PAGE_CLEAN;
	}
	return(0);
}
#endif NOTDEF

DbmRc
CacheDup PARAMS((oldcache, newcache),
PageCache *oldcache			_AND_
PageCache **newcache)
{
	PageCache *pc;

	if ((pc = (PageCache *) NAMEDMALLOC("PageCache", sizeof(PageCache)))
		== NULL)
		return(DBM_MALLOC_FAILED);
	if ((pc->page = (void *) NAMEDMALLOC("CacheDup", oldcache->nbytes))
		== NULL) {
		FREE(pc);
		return(DBM_MALLOC_FAILED);
	}

	pc->pagenum = oldcache->pagenum;
	pc->nbytes = oldcache->nbytes;
	pc->flags = oldcache->flags;
	pc->next = NULL;
	bcopy((char *) oldcache->page, (char *) pc->page, oldcache->nbytes);

	*newcache = pc;

	return(DBM_OK);
}
