static char *SCCS_Id = "@(#)complete.c\tver 1.1 (92/05/11 15:51:24)";

/*
 * Copyright 1992 Barry Brachman
 * See included file "copyright.h" for complete copyright information.
 */

#include <ctype.h>
#include <stdio.h>

#include "complete.h"

#ifdef __USE_ANSI_C__
int complete(char *, int, void *, int (*)(), int (*)(), Completions *);
int complete_compare(char *, char *, int *);

static int binary_search(char *, int, void *, int (*)());
#else
int complete();
int complete_compare();

static int binary_search();
#endif

/*
 * word        - the substring to be completed
 * nwords      - the number of words in the sorted word list
 * wordvec     - a vector of pointers to the word list
 * compar1     - comparison function (also returns length of matching prefix)
 * compar2     - alternate call to compar1()
 * completions - result of the call
 */
int
complete(word, nwords, wordvec, compar1, compar2, completions)
char *word;
int nwords;
void *wordvec;
int (*compar1)(), (*compar2)();
Completions *completions;
{
	register int i, st;
	int len, prefix_cnt;

	completions->start = -1;
	completions->end = -1;
	completions->common = 0;
	i = binary_search(word, nwords, wordvec, compar1);
	len = strlen(word);
	for (; i < nwords; i++) {
		st = (*compar1)(word, wordvec, i, &prefix_cnt);
		if (prefix_cnt == len) {
			if (completions->start < 0)
				completions->start = completions->end = i;
			else
				completions->end = i;
		}
		else if (st < 0 || completions->start >= 0)
			break;
	}
	if (completions->start >= 0) {
		if (completions->end == completions->start)
			(*compar2)(wordvec, completions->start,
					   wordvec, completions->start,
					   &completions->common);
		else {
			for (i = 1; i <= completions->end - completions->start; i++) {
				(*compar2)(wordvec, completions->start,
						   wordvec, completions->start + i,
						   &completions->common);
			}
		}
	}
	return(completions->start < 0
		   ? 0 : completions->end - completions->start + 1);
}

/*
 * Binary search
 */
static int
binary_search(word, nwords, wordvec, compar)
char *word;
int nwords;
void *wordvec;
int (*compar)();
{
    register int low, high, mid, cond;
	int prefix_cnt;

    low = 0;
    high = nwords - 1;
	mid = 0;
    while (low <= high) {
        mid = (low + high) / 2;
        if ((cond = (*compar)(word, wordvec, mid, &prefix_cnt)) < 0)
            high = mid - 1;
        else if (cond > 0)
            low = mid + 1;
        else
            break;
    }
    return(mid);
}

int
complete_compare(p, s, prefix_cnt)
register char *p, *s;
int *prefix_cnt;
{
	register int ch1, ch2;

	*prefix_cnt = 0;
	while (1) {
		ch1 = isupper(*p) ? tolower(*p) : *p;
		ch2 = isupper(*s) ? tolower(*s) : *s;
		if (ch1 != ch2)
			return(ch1 - ch2);
		if (*p++ == '\0')
			return(0);
		s++;
		(*prefix_cnt)++;
	}
}

#ifdef TESTING

static int compar1();
static int compar2();
static int qcompar();

main(argc, argv)
int argc;
char **argv;
{
	FILE *fp;
	int i, len, nwords;
	char buf[BUFSIZ], **words;

	if (argc == 2) {
		fp = fopen(argv[1], "r");
		words = (char **) malloc(30000);
		for (nwords = 0; nwords < 30000; nwords++) {
			if (fgets(buf, sizeof(buf), fp) == NULL)
				break;
			len = strlen(buf);
			buf[len - 1] = '\0';
			words[nwords] = (char *) malloc(len + 1);
			strcpy(words[nwords], buf);
		}
		fclose(fp);
	}
	else
		exit(1);

	qsort(words, nwords, sizeof(char *), qcompar);

	while (1) {
		Completions c;
		int compar1(), compar2();

		printf("word? ");
		gets(buf);
		printf("return = %d\n",
			   complete(buf, nwords, words, compar1, compar2, &c));
		if (c.start < 0)
			printf("No matches.\n");
		else {
			for (i = c.start; i <= c.end; i++)
				printf("%s\n", words[i]);
			printf("Common: %d\n", c.common);
		}
	}
}

static int
compar1(word, str, str_offset, prefix_cnt)
char *word, **str;
int str_offset, *prefix_cnt;
{
	register char *p, *s;

	p = word;
	s = str[str_offset];
	*prefix_cnt = 0;
	while (*p == *s++) {
		if (*p++ == '\0')
			return(0);
		(*prefix_cnt)++;
	}
	return((unsigned) *p - (unsigned) *(s - 1));
}

static int
compar2(str1, str1_offset, str2, str2_offset, prefix_cnt)
char **str1, **str2;
int str1_offset, str2_offset, *prefix_cnt;
{

	return(compar1(str1[str1_offset], str2, str2_offset, prefix_cnt));
}

static int
qcompar(p, q)
char **p, **q;
{

	return(strcmp(*p, *q));
}
#endif TESTING
