static char *SCCS_Id = "@(#)error.c\tver 1.2 (92/08/25 15:17:14)";

/*
 * Copyright 1992 Barry Brachman
 * See included file "copyright.h" for complete copyright information.
 */

#include <stdio.h>

#include "local.h"

#define OK			DBM_NOERROR
#define WARNING     DBM_WARNING
#define INTERNAL	DBM_INTERNAL
#define FATAL		DBM_FATAL
#define BADNEWS     DBM_BADNEWS
#define CORRUPTED	DBM_CORRUPTED

static struct error {
	DbmRc         rc;
	DbmErrorClass class;
	char          *msg;
} dbmerror[] = {
	DBM_OK,					OK,        "(no error)",
	DBM_OPEN_FAILED,  		CORRUPTED, "Open failed",
	DBM_CLOSE_FAILED,       CORRUPTED, "Close failed",
	DBM_KEY_TOO_LONG,       WARNING,   "Key is too long",
	DBM_ENTRY_NOT_FOUND,    WARNING,   "Entry not found",
	DBM_PAGE_FULL,          INTERNAL,  "Page full",
	DBM_PAGE_NOT_FOUND,     INTERNAL,  "Page not found",
	DBM_PAGE_IO_ERROR,      CORRUPTED, "Page I/O error",
	DBM_PAGE_HASH_OVERFLOW, WARNING,   "Page hash overflow",
	DBM_PAGE_ALLOC_FAILED,  WARNING,   "Page allocation failed",
	DBM_ENTRY_EXISTS,       WARNING,   "Entry exists (insert mode specified)",
	DBM_UNIMPLEMENTED,      WARNING,   "Unimplemented operation required",
	DBM_SPLITMAP_TOO_BIG,   WARNING,   "Splitmap has grown too big",
	DBM_BAD_ALIGNMENT,      WARNING,   "Illegal alignment",
	DBM_PAGE_FREE_FAILED,   CORRUPTED, "Page free failed",
	DBM_NULL_KEY,           WARNING,   "A key of length zero was passed",
	DBM_PAGE_MUST_SPLIT,    INTERNAL,  "No room, must split page",
	DBM_PAGE_MAPPING_ERROR, INTERNAL,  "Page mapping error",
	DBM_PAGENUMBER_TOO_BIG,	WARNING,   "Page number is too big",
	DBM_PAGE_HASNT_SPLIT,  	INTERNAL,  "Page has not split",
	DBM_PAGE_HAS_SPLIT, 	INTERNAL,  "Page has split",
	DBM_PAGE_NO_PHYSMAP,	WARNING,   "Page physical page mapping",
	DBM_PAGE_EXISTS,   		INTERNAL,  "Page exists",
	DBM_PAGE_CORRUPT,		FATAL,     "Page has been corrupted",
	DBM_MALLOC_FAILED, 		FATAL,     "Malloc failed",
	DBM_TRANS_IN_PROGRESS,  WARNING,   "A transaction is in progress",
	DBM_NO_TRANSACTION, 	WARNING,   "There is no current transaction",
	DBM_RECOVER_IO_ERROR,	CORRUPTED, "Recovery failed - I/O error",
	DBM_RECOVER_FAILED,		CORRUPTED, "Recovery failed",
	DBM_BAD_RECOVER_FILE,	WARNING,   "Orphaned transaction file found",
	DBM_FLOCK_FAILED, 		FATAL,     "File locking failed",
	DBM_UNLINK_FAILED, 		WARNING,   "Could not unlink",
	DBM_FSYNC_FAILED,  		CORRUPTED, "Fsync() failed",
	DBM_TRANS_IO_ERROR,		CORRUPTED, "I/O error during top-level commit",
	DBM_PAGE_SPLIT_FAILED,  BADNEWS,   "Could not split page",
	DBM_PAGESIZE_TOO_SMALL,	FATAL,     "Config pagesize is too small",
	DBM_TRANS_LOCK_REQUEST_FAILED, BADNEWS, "Lock operation failed",
	DBM_TRANS_LOCK_PROTOCOL_ERROR, BADNEWS, "Unexpected lock state found",
	DBM_INVALID_ARGUMENT,   WARNING,   "An invalid argument was passed",
	DBM_ITER_NOT_INITIALIZED, WARNING, "Iteration must be (re)initialized",
	DBM_TRANS_ABORTED,      WARNING,
	    "The current transaction has been aborted by the system",
	DBM_NO_FREEMAP_SPACE,   WARNING,   "Insufficient free space in dbm",
	DBM_TRANS_PRECOMMITTED, WARNING,   "Transaction has been precommitted",
	DBM_TRANS_NOT_TOPLEVEL, WARNING,   "Must be top-level transaction",
	DBM_UNIX_ERROR,			FATAL,
	    "Unix system call or library routine failed",
	DBM_RECOVER_NO_DISK_SPACE, WARNING,
		"Out of disk space - free some and recommit transaction",
	DBM_OK,					OK, NULL
};

static char errorbuf[BUFSIZ];

char *
DbmErrorString PARAMS((rc),
DbmRc rc)
{
	int i;
	char *p, *q;

	for (i = 0; dbmerror[i].msg; i++) {
		if (dbmerror[i].rc == rc) {
			switch (dbmerror[i].class) {
			case OK:
				q = NULL;
				break;
			case WARNING:
				q = " (R)";
				break;
			case INTERNAL:
				q = " (I)";
				break;
			case CORRUPTED:
				q = " (C)";
				break;
			case BADNEWS:
				q = " (B)";
				break;
			case FATAL:
				q = " (F)";
				break;
			default:
				q = " (?)";
				break;
			}
			p = errorbuf;
			sprintf(p, "%s. %s", dbmerror[i].msg, q ? q : "");
			return(errorbuf);
		}
	}
	return("(DbmErrorString: Invalid rc!)");
}

DbmErrorClass
DbmRcClass PARAMS((rc),
DbmRc rc)
{
	int i;

	for (i = 0; dbmerror[i].msg; i++) {
		if (dbmerror[i].rc == rc)
			return(dbmerror[i].class);
	}
	return(BADNEWS);
}
