static char *SCCS_Id = "@(#)lib.c\tver 1.2 (92/08/25 15:17:41)";

/*
 * Copyright 1992 Barry Brachman
 * See included file "copyright.h" for complete copyright information.
 */

#include "local.h"

#ifdef THREADS

/*
 * Duplicate Unix strdup functionality
 */
char *
Strdup PARAMS((p),
char *p)
{
    char *newp;

    if ((newp = (char *) NAMEDMALLOC("Strdup", strlen(p) + 1)) != NULL)
        strcpy(newp, p);
    return(newp);
}

#else

typedef enum {
	LOCK_AVAILSTATE, LOCK_UNLOCKEDSTATE, LOCK_READSTATE, LOCK_WRITESTATE
} LOCKSTATE;

static LOCKSTATE locks[NLOCKS];

int
FreeLock PARAMS((locknum),
int locknum)
{

	if (locknum < 0 || locknum >= NLOCKS)
		return(-1);

	if (locks[locknum] != LOCK_UNLOCKEDSTATE)
		return(-1);

	locks[locknum] = LOCK_AVAILSTATE;
	return(0);
}

int
NewLock NOPARAMS()
{
	int found_pos, start_pos;
	static int pos;
	static int done_init = 0;

	if (!done_init) {
		for (pos = 0; pos < NLOCKS; pos++)
			locks[pos] = LOCK_AVAILSTATE;
		pos = 0;
		done_init = 1;
	}

	start_pos = pos;
	do {
		if (locks[pos] == LOCK_AVAILSTATE) {
			locks[pos] = LOCK_UNLOCKEDSTATE;
			found_pos = pos;
			if (++pos == NLOCKS)
				pos = 0;
			return(found_pos);
		}
		if (++pos == NLOCKS)
			pos = 0;
	} while (pos != start_pos);

	return(-1);
}

int
ReadLock PARAMS((locknum, func, arg),
int locknum			_AND_
int (*func)()		_AND_
void *arg)
{

	if (locknum < 0 || locknum >= NLOCKS)
		return(-1);

	if (func) {
		if ((*func)(arg, LOCK_READMODE) <= 0)
			return(-1);
	}

	if (locks[locknum] == LOCK_WRITESTATE)
		return(-1);

	locks[locknum] = LOCK_READSTATE;
	return(0);
}

int
SeekNRead PARAMS((fd, buf, len, offset, mode),
int fd				_AND_
char *buf			_AND_
int len				_AND_
off_t offset		_AND_
int mode)
{

	if (lseek(fd, offset, mode) < 0)
		return(-1);
	return(read(fd, buf, len));
}

int
SeekNWrite PARAMS((fd, buf, len, offset, mode),
int fd				_AND_
char *buf			_AND_
int len				_AND_
off_t offset		_AND_
int mode)
{

	if (lseek(fd, offset, mode) < 0)
		return(-1);
	return(write(fd, buf, len));
}

void
ShowLocks NOPARAMS()
{
	int pos;

	for (pos = 0; pos < NLOCKS; pos++) {
		if (locks[pos] != LOCK_AVAILSTATE)
			printf("%3d: %s\n",
				   pos,
				   (locks[pos] == LOCK_UNLOCKEDSTATE ? "Unlocked"
					: ((locks[pos] == LOCK_READSTATE) ? "Readlocked"
					: ((locks[pos] == LOCK_WRITESTATE) ? "Writelocked"
					: "???"))));
	}
}

int
Unlock PARAMS((locknum),
int locknum)
{

	if (locknum < 0 || locknum >= NLOCKS)
		return(-1);

	if (locks[locknum] != LOCK_READSTATE && locks[locknum] != LOCK_WRITESTATE)
		return(-1);

	locks[locknum] = LOCK_UNLOCKEDSTATE;
	return(0);
}

int
UpgradeLock PARAMS((locknum, func, arg),
int locknum			_AND_
int (*func)()		_AND_
void *arg)
{

	if (locknum < 0 || locknum >= NLOCKS)
		return(-1);

	if (func) {
		if ((*func)(arg, LOCK_WRITEMODE) <= 0)
			return(-1);
	}

	if (locks[locknum] != LOCK_READSTATE)
		return(-1);

	locks[locknum] = LOCK_WRITESTATE;
	return(0);
}

int
WriteLock PARAMS((locknum, func, arg),
int locknum			_AND_
int (*func)()		_AND_
void *arg)
{

	if (locknum < 0 || locknum >= NLOCKS)
		return(-1);

	if (func) {
		if ((*func)(arg, LOCK_WRITEMODE) <= 0)
			return(-1);
	}

	if (locks[locknum] != LOCK_UNLOCKEDSTATE)
		return(-1);

	locks[locknum] = LOCK_WRITESTATE;
	return(0);
}

#ifndef sun
char *
strdup PARAMS((s),
char *s)
{
	char *new;

	if (new = (char *) malloc(strlen(s) + 1))
		strcpy(new, s);
	return(new);
}
#endif

#endif

