/* SCCS_Id = "@(#)local.h\tver 1.3 (93/03/15 13:52:40)" */

#ifndef _DBM_LOCAL_H_
#define _DBM_LOCAL_H_

#include <sys/types.h>
#ifdef mips
#include "/usr/include/unistd.h"
#else
#include <unistd.h>
#endif
#include <stdio.h>

#include "ansi.h"
#include "tdbm.h"
#include "threads.h"
#include "copyright.h"

#define STREQ(P, Q)					(!(strcmp((P), (Q))))
#define OPTION(P, OPTION, DEFAULT)	(((P) && (P->OPTION)) ? \
										(P->OPTION) : (DEFAULT))

/* Defaults */
#ifndef DBM_DEFAULT_MODE
#define DBM_DEFAULT_MODE			0644
#endif

#ifndef DBM_DEFAULT_PAGESIZE
#define DBM_DEFAULT_PAGESIZE		8192
#endif

#ifndef DBM_DEFAULT_ALLOC_UNITS
#define DBM_DEFAULT_ALLOC_UNITS		1
#endif

#define DBM_DEFAULT_TRANSDIR		"./"

/* Define LOCK_WARN to enable printing of warning message if dbm unlockable. */
/* #define LOCK_WARN		1 */

/* For transactions: */
#define TRANS_COMMIT_FLAG			-1
#define TRANS_MAGIC_FLAG			0x17171717	/* The most random flag. */
#define TRANS_EOL_FLAG				-1
#define TRANS_FILE_SUFFIX			".trans"
#define TRANS_FILE_RE_EXP			".*"
#define TRANS_FILE_DATE_LEN			10
#define TRANS_FILE_MIN_DIRSIZE		BUFSIZ
/* The following two modes must be different. */
#define TRANS_FILE_CREATE_MODE		0700
#define TRANS_FILE_PRECOMMIT_MODE	0600
#define TRANS_FILE_COMMIT_MODE		0400

#define DBM(trans)					((trans)->dbm)
#define TID(trans)					((trans)->tid)
#define DBF(trans)					((trans)->dbf)

typedef struct DbmTrans {
	Dbm		*dbm;
	DbmFile *dbf;
	Tid		*tid;
} DbmTrans;

/* Alignment required for the Indirect structure. */
#define INDIRECT_ALIGN		ALIGN4

typedef struct {
	PhysPagenum	pagenum;
	int			length;
} Indirect;

typedef enum { MUST_EXIST, MAY_CREATE } ReadMode;

typedef enum { ABORTING, COMMITTING } TransResult;

/* alloc.c */
extern DbmRc AllocBytes PROTO((DbmTrans *trans, int nbytes, PhysPagenum *start));
extern DbmRc FreeBytes PROTO((DbmTrans *trans, PhysPagenum first_page, int nbytes));
extern DbmRc PrintFreemap PROTO((Dbm *dbm, Tid *tid, int dirpagenum));

/* cache.c */
extern DbmRc CacheAlloc PROTO((PhysPagenum pagenum, int nbytes, PageCache **cache));
extern DbmRc CacheDup PROTO((PageCache *oldcache, PageCache **newcache));
extern void CacheFree PROTO((PageCache *cache));
extern DbmRc CacheRead PROTO((DbmTrans *trans, PhysPagenum physpage, int nbytes,
					   PageCache **cache));
extern DbmRc CacheWrite PROTO((DbmTrans *trans, PageCache *cache));

/* entry.c */
extern DbmRc StoreEntry PROTO((DbmTrans *trans, Datum *key, Datum *value, DbmStoreMode mode));
extern DbmRc DeleteEntry PROTO((DbmTrans *trans, Datum *key));
extern DbmRc GetFirstKey PROTO((DbmTrans *trans, Datum *key));
extern DbmRc GetNextKey PROTO((DbmTrans *trans, Datum *key));
extern DbmRc GetValue PROTO((DbmTrans *trans, Datum *key, Datum *value, int getvalue));
extern int MaxDirectEntrySize PROTO((DbmFile *dbf));
extern DbmRc PrintPage PROTO((Dbm *dbm, Tid *tid, LogPagenum pagenum));
extern DbmRc PrintUtilization PROTO((Dbm *dbm, Tid *tid));

/* lib.c */
extern int breakup PROTO((char *, char ***, char **, int *, char *));
extern int breakupq PROTO((char *, char ***, char **, int *, char *, int));
extern int mkargvec PROTO((char *, int *, char **, char *, char *));
extern int mkargvecq PROTO((char *, int *, char **, char *, char *, int));

/* open.c */
extern void PrintHeaders PROTO((Tid *tid));
extern void PrintOpen PROTO((void));
extern void RemoveFromDbflist PROTO((DbmFile *dbf));
extern void ShowDbm PROTO((Dbm *dbm));

/* page.c */
extern DbmRc AddIndirect PROTO((DbmTrans *trans, Datum *value, Indirect *indirect));
extern DbmRc DeleteIndirect PROTO((DbmTrans *trans, Indirect *indirect));
extern DbmRc DeletePage PROTO((DbmTrans *trans, LogPagenum pagenum));
extern int Hashinc PROTO((int hash, int mask));
extern DbmRc HashToPagenum PROTO((DbmTrans *trans, int hashval, int *maskp,
						   LogPagenum *logpagenum));
extern DbmRc KeyToPagenum PROTO((DbmTrans *trans, Datum *key, LogPagenum *pagenum));
extern DbmRc NewLogPage PROTO((DbmTrans *trans, LogPagenum pagenum, int npages));
extern DbmRc PrintPhysmap PROTO((Dbm *dbm, Tid *tid, int dirpagenum));
extern DbmRc PrintSplitmap PROTO((Dbm *dbm, Tid *tid, int dirpagenum));
extern DbmRc ReadIndirect PROTO((DbmTrans *trans, PhysPagenum pagenum, Datum *value));
extern DbmRc ReadPage PROTO((DbmTrans *trans, PageCache **cache, LogPagenum pagenum,
					  ReadMode mode));
extern DbmRc SetSplitmapBit PROTO((DbmTrans *trans, LogPagenum pagenum));
extern DbmRc SplitPagenum PROTO((DbmTrans *trans, Datum *key, LogPagenum *logpagenum));
extern DbmRc UpdatePage PROTO((DbmTrans *trans, PageCache **cachep));
extern DbmRc WriteIndirect PROTO((DbmTrans *trans, PhysPagenum pagenum, Datum *value));
extern DbmRc WritePage PROTO((DbmTrans *trans, PageCache *cache));

/* trans.c */
extern void PrintActiveTrans PROTO((Tid *tid));
extern void ShowDbfLocks PROTO((void));
extern void ShowTransLocks PROTO((Tid *tid));
extern DbmRc TransAbort PROTO((Tid *tid));
extern DbmRc TransBegin PROTO((Tid *parent, Tid **child));
extern DbmRc TransCommit PROTO((Tid *tid));
extern DbmRc TransDelete PROTO((DbmTrans *trans, PhysPagenum pagenum));
extern DbmRc TransPrecommit PROTO((Tid *tid, DbmTidName *tidname));
extern DbmRc TransRead PROTO((DbmTrans *trans, PhysPagenum pagenum, int nbytes,
							  PageCache **cache));
extern DbmRc TransRecover PROTO((Dbm *dbm, DbmRecovery **rl));
extern DbmRc TransTerminate PROTO((Dbm *dbm, DbmTerminateMode mode));
extern DbmRc TransUncache PROTO((DbmFile *dbf));
extern DbmRc TransUpdate PROTO((DbmTrans *trans, PageCache **cachep));
extern DbmRc TransUserRecover PROTO((DbmFile *dbf, DbmRecovery **recoveryp,
									 Tid **tidp));
extern DbmRc TransWrite PROTO((DbmTrans *trans, PageCache *cache));

#endif
