static char *SCCS_Id = "@(#)tdbm.c\tver 1.5 (94/07/06 11:02:37)";

/*
 * Copyright 1992 Barry Brachman
 * See included file "copyright.h" for complete copyright information.
 */

#include <sys/types.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <errno.h>
#include <stdio.h>

#include "tdbm.h"
#include "threads.h"
#include "complete.h"

#define STREQ(P, Q)			(!strcmp(P, Q))

typedef struct {
	char *name;
	int arg_count;
	void (*command) PROTO((int argc, char **argv));
} Command;

typedef struct TidTable {
	char			*name;
	Tid				*tid;
	struct TidTable	*parent;
} TidTable;

static int compar1 PROTO((char *word, Command *str, int str_offset,
					int *prefix_cnt));
static int compar2 PROTO((Command *str1, int str1_offset, Command *str2,
				   int str2_offset, int *prefix_cnt));
static void do_abort PROTO((int argc, char **argv));
static void do_align PROTO((int argc, char **argv));
static void do_begin PROTO((int argc, char **argv));
static void do_close PROTO((int argc, char **argv));
static void do_commit PROTO((int argc, char **argv));
static void do_delete PROTO((int argc, char **argv));
static void do_echo PROTO((int argc, char **argv));
static void do_fetch PROTO((int argc, char **argv));
static void do_first PROTO((int argc, char **argv));
static void do_help PROTO((int argc, char **argv));
static void do_look PROTO((int argc, char **argv));
static void do_next PROTO((int argc, char **argv));
static void do_open PROTO((int argc, char **argv));
static void do_precommit PROTO((int argc, char **argv));
static void do_print PROTO((int argc, char **argv));
static void do_quit PROTO((int argc, char **argv));
static void do_showtrans PROTO((int argc, char **argv));
static void do_source PROTO((int argc, char **argv));
static void do_system PROTO((int argc, char **argv));
static void do_trans PROTO((int argc, char **argv));
static void do_update PROTO((int argc, char **argv));
static void do_zap PROTO((int argc, char **argv));
static int loadFile PROTO((char *pathname, Datum *value));
static TidTable *lookupTid PROTO((char *name));
static TidTable *newTid PROTO((void));
static void prog PROTO((int arg));
static void usage PROTO((void));

static Command command_tab[] = {
	/* XXX These must be sorted by command name. */
	"!",         -1, do_system,
	".",		  1, do_source,
	"?",	  	 -1, do_help,
	"abort",	 -1, do_abort,
	"align",	 -1, do_align,
	"begin",	 -1, do_begin,
	"close",	 -1, do_close,
	"commit",	  0, do_commit,
	"delete",	 -1, do_delete,
	"echo",      -1, do_echo,
	"fetch",	 -1, do_fetch,
	"first",	  0, do_first,
	"help",		 -1, do_help,
	"insert",	 -1, do_update,
	"look",		 -1, do_look,
	"next",		  0, do_next,
	"open",		 -1, do_open,
	"precommit",  0, do_precommit,
	"print",	 -1, do_print,
	"quit",		  0, do_quit,
	"replace",	 -1, do_update,
	"source",	  1, do_source,
	"trans",	  1, do_trans,
	"where",      0, do_showtrans,
	"zap",		  0, do_zap
};

#define NTIDS		10
static TidTable tidtab[NTIDS];
static TidTable *current_tid = NULL;
static char cmd[BUFSIZ], prevcmd[BUFSIZ];
static int aborting, align;
static Dbm *dbm;
static FILE *fpin;

extern char *index PROTO((char *, char));
extern char *sys_errlist[];
extern int errno;

int ARGC;
char **ARGV;

#ifdef THREADS
mainp PARAMS((argc, argv),
int argc			_AND_
char **argv)
{

	ARGC = argc;
	ARGV = argv;
	Create(prog, 32767, "Prog", NULL, LOW);
}
#else
main PARAMS((argc, argv),
int argc			_AND_
char **argv)
{

	ARGC = argc;
	ARGV = argv;
	prog(NULL);
}
#endif

static void
prog PARAMS((arg),
int arg)
{
	int argc, i, mkargc, n, st;
	char **argv, **mkargv, *space;
	Command *command;
	Completions completions;
	DbmRc rc;

	argc = ARGC;
	argv = ARGV;

	dbm = NULL;
	aborting = 0;
	align = 0;
	fpin = stdin;
	prevcmd[0] = '\0';
	for (i = 0; i < NTIDS; i++) {
		tidtab[i].name = NULL;
		tidtab[i].parent = NULL;
		tidtab[i].tid = NULL;
	}

	while (1) {
		bzero(cmd, sizeof(cmd));
		if (fpin == stdin) {
			printf("? ");
			fflush(stdout);
		}
		if (fgets(cmd, BUFSIZ, fpin) == NULL) {
			if (fpin == stdin)
				break;
			fclose(fpin);
			fpin = stdin;
			continue;
		}

		if (cmd[0] == '#')
			continue;
		if (cmd[0] == '\n') {
			if (fpin == stdin)
				strcpy(cmd, prevcmd);
			else
				continue;
		}
		else {
			cmd[strlen(cmd) - 1] = '\0';
			strcpy(prevcmd, cmd);
		}

		if (breakupq(cmd, &mkargv, &space, &mkargc, NULL, 0) < 0) {
			printf("???\n");
			continue;
		}
		if (mkargc == 0) {
			FREE(space);
			FREE(mkargv);
			continue;
		}

		n = complete(mkargv[0], sizeof(command_tab) / sizeof(Command),
					 command_tab, compar1, compar2, &completions);

		st = -1;
		if (n == 1
			|| (n > 1
				&& STREQ(mkargv[0], command_tab[completions.start].name))) {
			command = &command_tab[completions.start];
#ifdef NOTDEF
			if (aborting && cmd[0] != 'B' && cmd[0] != 'c' && cmd[0] != 'z') {
				if (fpin == stdin)
					printf("(still aborting)\n");
				continue;
			}
#endif NOTDEF
			aborting = 0;

			if (command->arg_count >= 0) {
				if (mkargc < command->arg_count + 1)
					printf("Missing arguments.\n");
				else if (mkargc > command->arg_count + 1)
					printf("Too many arguments.\n");
				else
					st = 0;
			}
			else
				st = 0;
		}
		else if (n > 1)
			printf("Ambiguous command.\n");
		else {
			printf("Unknown command.\n");
			usage();
		}
		if (!st)
			(*command->command)(mkargc, mkargv);

		if (aborting == 1) {
			fprintf(stderr, "Aborting.\n");
			if ((rc = DbmAbort(tidtab[0].tid)) != DBM_OK)
				fprintf(stderr, "%s\n", DbmErrorString(rc));
			tidtab[0].tid = NULL;
			aborting = 2;
		}
		FREE(mkargv);
		FREE(space);
	}
	exit(0);
}

static int
compar1 PARAMS((word, str, str_offset, prefix_cnt),
char *word				_AND_
Command *str			_AND_
int str_offset			_AND_
int *prefix_cnt)
{

    return(complete_compare(word, str[str_offset].name, prefix_cnt));
}

static int
compar2 PARAMS((str1, str1_offset, str2, str2_offset, prefix_cnt),
Command *str1			_AND_
int str1_offset			_AND_
Command *str2			_AND_
int str2_offset			_AND_
int *prefix_cnt)
{

    return(complete_compare(str1[str1_offset].name, str2[str2_offset].name, prefix_cnt));
}

static void
do_abort PARAMS((argc, argv),
int argc			_AND_
char **argv)
{
	DbmRc rc;

	if (current_tid == NULL) {
		fprintf(stderr, "No such transaction.\n");
		return;
	}
	if ((rc = DbmAbort(current_tid->tid)) != DBM_OK)
		fprintf(stderr, "%s\n", DbmErrorString(rc));
	else {
		FREE(current_tid->name);
		current_tid->name = NULL;
		current_tid = current_tid->parent;
	}
	if (current_tid && fpin == stdin)
		printf("Current transaction: %s\n", current_tid->name);
}

static void
do_align PARAMS((argc, argv),
int argc			_AND_
char **argv)
{

	if (fpin == stdin)
		printf("Was %d\n", align);
	align = atoi(argv[1]);
}

static void
do_begin PARAMS((argc, argv),
int argc			_AND_
char **argv)
{
	DbmRc rc;
	TidTable *newtidp, *parent;
	char buf[10], *name;
	static int count = 1;

	if (argc > 2) {
		fprintf(stderr, "Bad arg count.\n");
		return;
	}
	if (!dbm) {
		fprintf(stderr, "Not opened.\n");
		return;
	}
	if (argc == 1) {
		do {
			sprintf(buf, "T%d", count++);
		} while (lookupTid(buf));
		name = buf;
	}
	else {
		name = argv[1];
		if (lookupTid(name)) {
			fprintf(stderr, "Name already in use.\n");
			return;
		}
	}
	
	if ((newtidp = newTid()) == NULL) {
		fprintf(stderr, "Out of tids.\n");
		return;
	}
	parent = current_tid;
	if ((rc = DbmBegin(parent ? parent->tid : NULL, &newtidp->tid)) != DBM_OK)
		fprintf(stderr, "%s\n", DbmErrorString(rc));
	else {
		newtidp->name = STRDUP(name);
		newtidp->parent = parent;
		current_tid = newtidp;
	}
}

static void
do_close PARAMS((argc, argv),
int argc			_AND_
char **argv)
{
	DbmRc rc;

	if (!dbm) {
		fprintf(stderr, "Not opened.\n");
		return;
	}
	if (current_tid)
		fprintf(stderr, "There is an uncommitted transaction.\n");
	else {
		if ((rc = DbmClose(dbm)) != DBM_OK)
			fprintf(stderr, "%s\n", DbmErrorString(rc));
		else
			dbm = NULL;
	}
}

static void
do_commit PARAMS((argc, argv),
int argc			_AND_
char **argv)
{
	DbmRc rc;

	if (current_tid) {
		if ((rc = DbmCommit(current_tid->tid)) != DBM_OK)
			fprintf(stderr, "%s\n", DbmErrorString(rc));
		else {
			FREE(current_tid->name);
			current_tid->name = NULL;
			current_tid = current_tid->parent;
		}
		if (current_tid && fpin == stdin)
			printf("Current transaction: %s\n", current_tid->name);
	}
	else
		printf("No current transaction.\n");
}

static void
do_delete PARAMS((argc, argv),
int argc			_AND_
char **argv)
{
	Datum key;
	DbmRc rc;

	if (!current_tid) {
		fprintf(stderr, "No transaction.\n");
		return;
	}
	key.dptr = argv[1];
	key.dsize = strlen(key.dptr) + 1;
	if ((rc = DbmDelete(dbm, current_tid->tid, key)) != DBM_OK) {
		fprintf(stderr, "'%s': %s\n", key.dptr, DbmErrorString(rc));
		aborting = 1;
	}
}

static void
do_echo PARAMS((argc, argv),
int argc			_AND_
char **argv)
{
	int i;

	for (i = 1; i < argc; i++)
		printf("%s\n", argv[i]);
}

static void
do_fetch PARAMS((argc, argv),
int argc			_AND_
char **argv)
{
	Datum key, value;
	DbmRc rc;
	FILE *fpout;

	if (argc < 2 || argc > 3) {
		fprintf(stderr, "Bad arg count.\n");
		return;
	}
	if (!current_tid) {
		fprintf(stderr, "No transaction.\n");
		return;
	}

	key.dptr = argv[1];
	key.dsize = strlen(key.dptr) + 1;
	key.desc = 0;
	if ((rc = DbmFetch(dbm, current_tid->tid, key, &value)) != DBM_OK)
		fprintf(stderr, "%s\n", DbmErrorString(rc));
	else {
		fpout = stdout;
		if (argc > 2 && argv[2][0] == '>') {
			if ((fpout = fopen(&argv[2][1], "w")) == NULL) {
				fprintf(stderr,
						"fopen: %s: %s\n", &argv[2][1], sys_errlist[errno]);
				fpout = stdout;
			}
		}
		fwrite(value.dptr, value.dsize, 1, fpout);
		if (fpout != stdout)
			fclose(fpout);
		printf("\n");
		printf("kdesc = %#o, dsize = %d, ddesc = %#o\n",
			   key.desc, value.dsize, value.desc);
	}
}

static void
do_first PARAMS((argc, argv),
int argc			_AND_
char **argv)
{
	Datum key, value;
	DbmRc rc;

	if (!current_tid) {
		fprintf(stderr, "No transaction.\n");
		return;
	}
	if ((rc = DbmFirstkey(dbm, current_tid->tid, &key)) != DBM_OK)
		fprintf(stderr, "%s\n", DbmErrorString(rc));
	else {
		printf("Key: '%s' (size=%d,desc=%#o)\n",
			   key.dptr, key.dsize, key.desc);
		if ((rc = DbmFind(dbm, current_tid->tid, key, &value)) != DBM_OK)
			fprintf(stderr, "%s\n", DbmErrorString(rc));
		else
			printf("Value: dsize = %d, ddesc = %#o\n",
				   value.dsize, value.desc);
	}
}

static void
do_help PARAMS((argc, argv),
int argc			_AND_
char **argv)
{

	usage();
}

static void
do_look PARAMS((argc, argv),
int argc			_AND_
char **argv)
{
	Datum key, value;
	DbmRc rc;

	if (!current_tid) {
		fprintf(stderr, "No transaction.\n");
		return;
	}
	key.desc = 0;
	key.dptr = argv[1];
	key.dsize = strlen(key.dptr) + 1;
	if ((rc = DbmFind(dbm, current_tid->tid, key, &value)) != DBM_OK)
		fprintf(stderr, "%s\n", DbmErrorString(rc));
	else
		printf("kdesc = %#o, dsize = %d, ddesc = %#o\n",
			   key.desc, value.dsize, value.desc);
}

static void
do_next PARAMS((argc, argv),
int argc			_AND_
char **argv)
{
	Datum key, value;
	DbmRc rc;

	if (!current_tid) {
		fprintf(stderr, "No transaction.\n");
		return;
	}
	if ((rc = DbmNextkey(dbm, current_tid->tid, &key)) != DBM_OK)
		fprintf(stderr, "%s\n", DbmErrorString(rc));
	else {
		printf("Key: '%s' (size=%d,desc=%#o)\n",
			   key.dptr, key.dsize, key.desc);
		if ((rc = DbmFind(dbm, current_tid->tid, key, &value)) != DBM_OK)
			fprintf(stderr, "%s\n", DbmErrorString(rc));
		else
			printf("Value: dsize = %d, ddesc = %#o\n",
				   value.dsize, value.desc);
	}
}

static void
do_open PARAMS((argc, argv),
int argc			_AND_
char **argv)
{
	char *name;
	DbmConfig config;
	DbmRc rc;

	if (dbm) {
		/* XXX Should commit/abort transactions. */
		if (current_tid)
			fprintf(stderr, "Warning: transaction was active.\n");
		if ((rc = DbmClose(dbm)) != DBM_OK)
			fprintf(stderr, "%s\n", DbmErrorString(rc));
	}
	config.mode = 0644;
	config.pagesize = 8192;
	config.allocunits = 2;

	if (argc == 1)
		name = "./db/gorp";
	else if (STREQ(argv[1], "-v")) {
		name = NULL;
		argv++;
		argc--;
	}
	else
		name = argv[1];
	rc = DbmOpen(name, name ? DBM_PERSISTENT : DBM_VOLATILE, &config, &dbm,
					NULL);
	if (rc != DBM_OK)
		fprintf(stderr, "%s\n", DbmErrorString(rc));
}

static void
do_precommit PARAMS((argc, argv),
int argc			_AND_
char **argv)
{
	DbmRc rc;

	if (current_tid) {
		if ((rc = DbmPrecommit(current_tid->tid, NULL)) != DBM_OK)
			fprintf(stderr, "%s\n", DbmErrorString(rc));
	}
	else
		printf("No current transaction.\n");
}

static void
do_print PARAMS((argc, argv),
int argc			_AND_
char **argv)
{
	int count, nbytes, ni, nibytes;
	Datum key, value;
	DbmRc rc;

	switch (argv[1][0]) {
	case 'e':
		do_echo(argc - 1, argv + 1);
		break;
	case 'f':
		if (!current_tid) {
			fprintf(stderr, "No transaction.\n");
			return;
		}
		count = (argc == 2 ? 0 : atoi(argv[2]));
		if ((rc = PrintFreemap(dbm, current_tid->tid, count)) != DBM_OK)
			fprintf(stderr, "%s\n", DbmErrorString(rc));
		break;
	case 'h':
		if (!current_tid) {
			fprintf(stderr, "No transaction.\n");
			return;
		}
		PrintHeaders(current_tid->tid);
		break;
	case 'k':
		if (!current_tid) {
			fprintf(stderr, "No transaction.\n");
			return;
		}
		count = 0;
		nbytes = 0;
		rc = DbmFirstkey(dbm, current_tid->tid, &key);
		if (rc == DBM_ENTRY_NOT_FOUND) {
			printf("Empty.\n");
			break;
		}
		while (rc == DBM_OK) {
			count++;
			nbytes += key.dsize;
			rc = DbmNextkey(dbm, current_tid->tid, &key);
		}
		if (rc != DBM_ENTRY_NOT_FOUND)
			fprintf(stderr, "%s\n", DbmErrorString(rc));
		printf("Count: %d, Bytes: %d\n", count, nbytes);
		break;
	case 'l':
		ShowLocks();
		break;
	case 'm':
		if (!current_tid) {
			fprintf(stderr, "No transaction.\n");
			return;
		}
		if ((rc = PrintPhysmap(dbm, current_tid->tid, atoi(argv[2])))
			!= DBM_OK)
			fprintf(stderr, "%s\n", DbmErrorString(rc));
		break;
	case 'o':
		PrintOpen();
		break;
	case 'p':
		if (!current_tid) {
			fprintf(stderr, "No transaction.\n");
			return;
		}
		count = (argc == 2 ? 0 : atoi(argv[2]));
		if ((rc = PrintPage(dbm, current_tid->tid, count)) != DBM_OK)
			fprintf(stderr, "%s\n", DbmErrorString(rc));
		break;
	case 's':
		if (!current_tid) {
			fprintf(stderr, "No transaction.\n");
			return;
		}
		count = (argc == 2 ? 0 : atoi(argv[2]));
		if ((rc = PrintSplitmap(dbm, current_tid->tid, count)) != DBM_OK)
			fprintf(stderr, "%s\n", DbmErrorString(rc));
		break;
	case 't':
		switch (argv[1][1]) {
		case '\0':
			do_showtrans(0, NULL);
			break;
		case 'f':
			ShowDbfLocks();
			break;
		case 't':
		if (!current_tid) {
			fprintf(stderr, "No transaction.\n");
			return;
		}
			ShowTransLocks(current_tid->tid);
			break;
		default:
			fprintf(stderr, "Unknown option flag.\n");
			break;
		}
		break;
	case 'u':
		if (!current_tid) {
			fprintf(stderr, "No transaction.\n");
			return;
		}
		if ((rc = PrintUtilization(dbm, current_tid->tid)) != DBM_OK)
			fprintf(stderr, "%s\n", DbmErrorString(rc));
		break;
	case 'v':
		if (!current_tid) {
			fprintf(stderr, "No transaction.\n");
			return;
		}
		count = 0;
		nbytes = 0;
		ni = 0;
		nibytes = 0;
		rc = DbmFirstkey(dbm, current_tid->tid, &key);
		if (rc == DBM_ENTRY_NOT_FOUND) {
			printf("Empty.\n");
			break;
		}
		while (rc == DBM_OK) {
			if ((rc = DbmFind(dbm, current_tid->tid, key, &value)) != DBM_OK)
				break;
			count++;
			nbytes += value.dsize;
			if (value.desc & INDIRECT) {
				ni++;
				nibytes += value.dsize;
			}
			rc = DbmNextkey(dbm, current_tid->tid, &key);
			if (rc == DBM_ENTRY_NOT_FOUND) {
				rc = DBM_OK;
				break;
			}
		}
		if (rc != DBM_OK) {
			fprintf(stderr, "%s\n", DbmErrorString(rc));
			fprintf(stderr, "Key: '%s' (size=%d,desc=%#o)\n",
					key.dptr, key.dsize, key.desc);
		}
		printf("Count: %d, Bytes: %d\n", count, nbytes);
		printf("Indirects: %d, Bytes: %d\n", ni, nibytes);
		break;

	default:
		break;
	}
}

static void
do_showtrans PARAMS((argc, argv),
int argc			_AND_
char **argv)
{
	int i;

	for (i = 0; i < NTIDS; i++) {
		if (tidtab[i].name) {
			printf("%c %s", &tidtab[i] == current_tid ? '*' : ' ',
				   tidtab[i].name);
			if (tidtab[i].parent)
				printf(" <- %s\n", tidtab[i].parent->name);
			else
				printf(" (Top)\n");
		}
	}
}

static void
do_quit PARAMS((argc, argv),
int argc			_AND_
char **argv)
{
	int i;
	DbmRc rc;

	if (dbm) {
		for (i = 0; i < NTIDS; i++) {
			if (tidtab[i].name) {
				if ((rc = DbmCommit(tidtab[i].tid)) != DBM_OK)
					fprintf(stderr, "%s\n", DbmErrorString(rc));
			}
		}
		if ((rc = DbmClose(dbm)) != DBM_OK)
			fprintf(stderr, "%s\n", DbmErrorString(rc));
	}
	exit(0);
}

static void
do_source PARAMS((argc, argv),
int argc			_AND_
char **argv)
{

	if (fpin == stdin) {
		if ((fpin = fopen(argv[1], "r")) == NULL) {
			fprintf(stderr, "Can't open '%s'\n", argv[1]);
			fpin = stdin;
		}
	}
	else
		fprintf(stderr, "Already reading from a file!\n");
}

static void
do_system PARAMS((argc, argv),
int argc			_AND_
char **argv)
{
	int i;
	char buf[BUFSIZ], *p;

	if (argc == 1)
		return;
	p = buf;
	for (i = 1; i < argc - 1; i++) {
		sprintf(p, "%s ", argv[i]);
		p += strlen(p);
	}
	sprintf(p, "%s", argv[argc - 1]);
	system(buf);
}

static void
do_trans PARAMS((argc, argv),
int argc			_AND_
char **argv)
{
	TidTable *tidp;

	if (tidp = lookupTid(argv[1]))
		current_tid = tidp;
	else
		printf("No such transaction.\n");
}

static void
do_update PARAMS((argc, argv),
int argc			_AND_
char **argv)
{
	Datum key, value;
	DbmRc rc;
	DbmStoreMode mode;

	if (argc < 2 || argc > 3) {
		fprintf(stderr, "Bad arg count.\n");
		return;
	}
	if (!dbm) {
		fprintf(stderr, "Not opened.\n");
		return;
	}
	if (!current_tid) {
		fprintf(stderr, "No transaction.\n");
		return;
	}
	key.dptr = argv[1];
	key.dsize = strlen(key.dptr) + 1;
	key.desc = 0;
	if (argv[2][0] == '<') {
		if (loadFile(&argv[2][1], &value) < 0)
			return;
		value.desc = 0;
	}
	else {
		value.dptr = argv[2];
		value.dsize = strlen(value.dptr) + 1;
		value.desc = align;
	}
	if (argv[0][0] == 'i')
		mode = DBM_INSERT;
	else
		mode = DBM_REPLACE;
	if ((rc = DbmStore(dbm, current_tid->tid, key, value, mode)) != DBM_OK) {
		fprintf(stderr, "%s\n", DbmErrorString(rc));
		aborting = 1;
	}
}

static void
do_zap PARAMS((argc, argv),
int argc			_AND_
char **argv)
{
	DbmRc rc;

	if (!dbm) {
		fprintf(stderr, "Not opened.\n");
		return;
	}
	if (current_tid) {
		if ((rc = DbmCommit(current_tid->tid)) != DBM_OK)
			fprintf(stderr, "%s\n", DbmErrorString(rc));
		current_tid->name = NULL;
		current_tid = NULL;
	}
	if (dbm->dbmfile->persistent && unlink(dbm->dbmfile->pathname) < 0)
		fprintf(stderr, "unlink: %s: %s\n",
				dbm->dbmfile->pathname, sys_errlist[errno]);
	if ((rc = DbmClose(dbm)) != DBM_OK)
		fprintf(stderr, "%s\n", DbmErrorString(rc));
	else
		dbm = NULL;
}

static int
loadFile PARAMS((pathname, value),
char *pathname		_AND_
Datum *value)
{
	int fd, rc;
	char *p;
	struct stat statbuf;

	rc = -1;
	p = NULL;
	if ((fd = OPEN(pathname, O_RDONLY, 0)) < 0) {
		fprintf(stderr, "open: %s: %s\n", pathname, sys_errlist[errno]);
		goto error;
	}
	if (fstat(fd, &statbuf) < 0) {
		fprintf(stderr, "fstat: %s: %s\n", pathname, sys_errlist[errno]);
		goto error;
	}
	if ((p = (char *) MALLOC(statbuf.st_size)) == NULL) {
		fprintf(stderr, "Malloc of %d bytes failed\n", statbuf.st_size);
		goto error;
	}
	if (READ(fd, p, statbuf.st_size) != statbuf.st_size) {
		fprintf(stderr, "read failed\n");
		goto error;
	}
	value->dptr = p;
	value->dsize = statbuf.st_size;

	rc = 0;
 error:
	if (fd >= 0)
		CLOSE(fd);
	if (rc && p)
		FREE(p);
	return(rc);
}

static TidTable *
newTid NOPARAMS()
{
	int i;

	for (i = 0; i < NTIDS; i++) {
		if (tidtab[i].name == NULL)
			return(&tidtab[i]);
	}
	return(NULL);
}

static TidTable *
lookupTid PARAMS((name),
char *name)
{
	int i;

	for (i = 0; i < NTIDS; i++) {
		if (tidtab[i].name && STREQ(name, tidtab[i].name))
			return(&tidtab[i]);
	}
	return(NULL);
}

static void
usage NOPARAMS()
{

	printf("This is tdbm %s\n", Version);
	printf("%s\n", Copyright);
	printf("Commands (word completion is done on commands):\n");
	printf("abort                    - abort current transaction\n");
	printf("align <#>                - set value alignment\n");
	printf("begin [<str>]            - begin transaction <str>\n");
	printf("close                    - close current dbm\n");
	printf("commit                   - commit current transaction\n");
	printf("delete <key>             - delete entry with <key>\n");
	printf("echo <text>              - print <text>\n");
	printf("fetch <key> [>file]      - fetch entry with <key>\n");
	printf("first                    - fetch key of first entry\n");
	printf("help                     - print this help message\n");
	printf("insert <key> [<]<value>  - insert <value> using <key>\n");
	printf("look <key>               - find (locate) entry with <key>\n");
	printf("next                     - fetch key of next entry\n");
	printf("open [file]              - open dbm <file> or default\n");
	printf("open -v                  - open a volatile dbm\n");
	printf("precommit                - precommit dbm\n");
	printf("print e <text>           - same as echo command\n");
	printf("      f #                - print freemap page #\n");
	printf("      h                  - print the header\n");
	printf("      k                  - print key statistics\n");
	printf("      l                  - print O/S locks\n");
	printf("      m #                - print physpagemap page #\n");
	printf("      o                  - print the control block for dbm\n");
	printf("      p #                - print logical page #\n");
	printf("      s #                - print splitmap page #\n");
	printf("      t[f|t]             - print active transactions\n");
	printf("      u                  - print utilization statistics\n");
	printf("      v                  - print value statistics\n");
	printf("quit                     - quit, closing dbm if necessary\n");
	printf("replace <key> [<]<value> - replace <value> for <key>\n");
	printf("source <file>            - source commands from <file>\n");
	printf("trans <str>              - make <str> the current transaction\n");
	printf("where                    - alias for 'print t'\n");
	printf("zap                      - close and remove dbm file\n");
	printf(". <file>                 - source commands from <file>\n");
	printf("#...                     - comment line\n");
	printf("?                        - print this help message\n");
	printf("! <shell command>        - shell escape\n");
	printf("<Return>                 - repeat last command\n");
}

static void err PROTO((void));

/*
 * Breakup a string into argc/argv format, allocating all the space
 * Assumes worst-case for number of pointers; i.e., an alternating
 * character/ifs sequence
 * Caller must free the space pool and the vector of pointers
 */
int
breakup PARAMS((buf, argv, sp, n, ifs),
char *buf			_AND_
char ***argv		_AND_
char **sp			_AND_
char *ifs			_AND_
int *n)
{
	register int len, rc;
	char **ptr;

	len = strlen(buf);
	if ((*sp = (char *) MALLOC((unsigned) (len + 1))) == NULL) {
		*n = 0;
		return(-4);
	}
	*n = (len + 1) / 2 + 1;
	ptr = (char **) MALLOC((unsigned) (sizeof(char *) * *n));
	if (ptr == NULL) {
		FREE(*sp);
		*n = 0;
		return(-4);
	}
	if ((rc = mkargvec(buf, n, ptr, *sp, ifs)) < 0) {
		FREE(*sp);
		FREE(ptr);
		*n = 0;
		return(rc);
	}
	*argv = ptr;
	return(0);
}

int
breakupq PARAMS((buf, argv, sp, n, ifs, keepq),
char *buf			_AND_
char ***argv		_AND_
char **sp			_AND_
char *ifs			_AND_
int *n				_AND_
int keepq)
{
	register int len, rc;
	char **ptr;

	len = strlen(buf);
	if ((*sp = (char *) MALLOC((unsigned) (len + 1))) == NULL) {
		*n = 0;
		return(-4);
	}
	*n = (len + 1) / 2 + 1;
	ptr = (char **) MALLOC((unsigned) (sizeof(char *) * *n));
	if (ptr == NULL) {
		FREE(*sp);
		*n = 0;
		return(-4);
	}
	if ((rc = mkargvecq(buf, n, ptr, *sp, ifs, keepq)) < 0) {
		FREE(*sp);
		FREE(ptr);
		*n = 0;
		return(rc);
	}
	*argv = ptr;
	return(0);
}

/*
 * mkargvec: scan a string and split up into argv
 * argc contains maximum number of argv elements upon entry
 * and is set to the actual number of elements on exit
 * Argv list terminated by entry of 0
 * If ifs is 0 or the empty string, then white space (tabs and blanks)
 * are used as field separators; otherwise each character in ifs
 * is considered to be a field separator.
 * Leading field separators in string are skipped.
 * Returns -1 if not enough argv elements, 0 oth.
 *
 * 22-Jun-86	Checking for no more argv pointers first caused problems when
 *		there were trailing blanks.  Delay the check until necessary.
 */

static char *default_ifs = "\t ";

int
mkargvec PARAMS((string, argc, argv, space, ifs),
char *string		_AND_
char *argv[]		_AND_
char *space			_AND_
char *ifs			_AND_
int *argc)
{
	register int i;
	register char *p, *sp;
	int max, count;

	p = string;
	sp = space;
	count = 0;
	max = *argc - 1;
	if (ifs == NULL || ifs[0] == '\0')
		ifs = default_ifs;

	/* For each arg, copy and set argv pointer */
	i = 0;
	while (1) {
		while (*p != '\0' && index(ifs, *p) != NULL)
			p++;
		if (*p == '\0')
			break;
		if (i == max)		/* no more pointers */
			break;
		argv[i++] = sp;
		count++;
		while (*p != '\0' && index(ifs, *p) == NULL)
			*sp++ = *p++;
		*sp++ = '\0';
	}
	*argc = count;
	argv[count] = 0;
	if (*p != '\0')			/* only if i == max */
		return(-1);
	else
		return(0);
}


/*
 * mkargvecq: modified mkargvec
 * scan a string and split up into argv
 * argc contains maximum number of argv elements upon entry
 * and is set to the actual number of elements on exit
 * Argv list terminated by entry of 0
 * If ifs is 0 or the empty string, then white space (tabs and blanks)
 * are used as field separators; otherwise each character in ifs
 * is considered to be a field separator.
 * Leading field separators in string are skipped.
 *
 * This version knows about quotes - single and double quotes may be used
 * to delimit words
 * Two consecutive quotes makes a null word
 * If the 'keepq' arg is non-zero, then the quote marks delimiting words are
 * kept
 *
 * Also knows about the escape character (backslash) which may be used to
 * prevent interpretation of ifs characters, quotes, or the escape char itself
 * Returns -1 if not enough argv elements, -2 if quotes screwed,
 * -3 if backslash quoting problem, 0 oth.
 */

#ifdef DEBUG
#include <stdio.h>
#endif DEBUG

int
mkargvecq PARAMS((string, argc, argv, space, ifs, keepq),
char *string		_AND_
char *argv[]		_AND_
char *space			_AND_
char *ifs			_AND_
int *argc			_AND_
int keepq)
{
	register int i;
	register char *p, *sp;
	int count, max;
	char *q, quote[10];

	p = string;
	sp = space;
	count = 0;
	max = *argc - 1;
	if (ifs == NULL || ifs[0] == '\0')
		ifs = default_ifs;

	quote[0] = '\0';
	if (index(ifs, '\'') == NULL)
		strcat(quote, "'");
	if (index(ifs, '"') == NULL)
		strcat(quote, "\"");

	if (quote[0] == '\0')
		err("Quote mode but quotes are all ifs");
	if (index(ifs, '\\') != NULL)
		err("Escape character (backslash) in ifs");

	q = NULL;
	/* For each arg, copy and set argv pointer */
	i = 0;
	while (1) {
		while (*p != '\0' && index(ifs, *p) != NULL)
			p++;
		if (*p == '\0')
			break;
		if (i == max)
			break;
		argv[i++] = sp;
		if ((q = index(quote, *p)) != NULL) {
			if (keepq)
				*sp++ = *q;
			if (*(++p) == '\0')
				break;
		}
		count++;
		while (*p != '\0') {
			if (q != NULL && *p == *q) {
				if (keepq)
					*sp++ = *q;
				q = NULL;
				p++;
				break;
			}
			else if (q == NULL && (q = index(quote, *p)) != NULL) {
				if (!keepq) {
					p++;
					continue;
				}
			}
			else if (*p == '\\') {
				if (*(++p) == '\0') {
					*argc = count;
					argv[count] = 0;
					err("Misquote");
					return(-3);
				}
				*sp++ = *p++;
				continue;
			}
			else if (q == NULL && index(ifs, *p) != NULL)
				break;
			*sp++ = *p++;
		}
		*sp++ = '\0';
	}
	*argc = count;
	argv[count] = 0;
	if (q != NULL) {
		err("Unmatched quote: %c", *q);
		return(-2);
	}

	if (*p != '\0') {		/* only if i == max */
		err("Not enough pointers");
		return(-1);
	}
	else
		return(0);
}

static void
err NOPARAMS()
{
}
